﻿using System;
using System.Windows.Forms;
using System.IO;
using System.Drawing;
using System.Collections.Generic;
using STFXmlParser;
using Selections;
using CTNet;
using PdfDocument;
using System.Text.RegularExpressions;
using System.Linq;

namespace CalcHelper
{
    public partial class Form1 : Form
    {
        /*
         * This routine handles Checkbox Control checkbox
         * lasso selections.
         * 
         * User can only lasso one controlled checkbox at a time.
         * User can not lasso both controlled and uncontrolled
         *      checkboxes at a time.
         *      
         * Called By:
         * 
         *  pictureBox1_MouseUp()
         *  pictureBox2_MouseUp()
         */
        public void ProcessLeftMouseUpChkBoxCtrl(ref Pdf pdf, ref PageArray pa)
        {
            /*
             * Caller has ensured that pdf, pa, and ctClient are not null,
             * and it is a left mouse button event.
            */
            if (pdf.mLassoMove)
            {
                bool lassoOK = true;
                List<Field> fa = null;
                List<string> nonCtrlSelList = new List<string>();
                List<string> ctrlSelList = new List<string>();
                List<string> ctrlUnSelList = new List<string>();
                int paIndex = (int)(pdf.curPage - 1);
                int boxIndex = -1;
                string errMsg = "";

                /*
                 * The lassoed checkboxes may end up in one of these
                 * selection lists:
                 * 
                 * A] Currently noncontrolled checkboxes to be selected
                 * B] Currently controlled checkboxes to be selected
                 * C] Currently controlled checkboxes to be unselected
                 * 
                 * While we are building the lists we will be looking for
                 * the below situations which will indicate a lasso error
                 * situation in which case we will inform the user and
                 * abort the lasso:
                 * 
                 * 1) User can only lasso one controlled checkbox at a
                 *      time. That one box can be either selected or
                 *      unselected.
                 * 2) User cannot lasso unselected uncontrolled checkboxes
                 *      along with controlled checkboxes.
                 */
                pa.GetFieldArrayRef(paIndex, ref fa);

                if (fa != null)
                {
                    try
                    {
                        for (int i = 0; i < fa.Count; i++)
                        {
                            if (fa[i].sbox.IntersectsWith(pdf.lassoRect))
                            {
                                Field field = fa[i];

                                if (field is TextField)
                                {
                                    if (((TextField)field).checkbox == true)
                                    {
                                        if (field.ctrlType == -1)
                                        {
                                            // Noncontrolled checkbox:
                                            if (field.leftBoxSel == false)
                                            {
                                                nonCtrlSelList.Add(field.ref_name);
                                            }
                                        }
                                        else if (field.ctrlType == docFile.chkBoxCtrl)
                                        {
                                            /*
                                             * Need to make sure we have a Checkbox
                                             * Controlled checkbox, not a dataflowed
                                             * checkbox, another type of a controlled
                                             * checkbox. Here we just ignore controlled
                                             * checkboxes other than Checkbox Controlled
                                             * ones since perhaps someone is doing a 
                                             * 'sloppy lasso'.
                                             */
                                            if (field.leftBoxSel == false)
                                            {
                                                // Selecting a controlled checkbox:
                                                ctrlSelList.Add(field.ref_name);
                                            }
                                            else
                                            {
                                                // Unselecting a controlled checkbox:
                                                ctrlUnSelList.Add(field.ref_name);
                                            }
                                        }

                                        // Check for a bad selection:
                                        if ((ctrlSelList.Count + ctrlUnSelList.Count) > 1)
                                        {
                                            errMsg = "Bad Checkbox Lasso:\n\nYou can only lasso one Checkbox Controlled checkbox [the colored checkboxes] at a time.\n\nAborting selection.";
                                        }
                                        else if ((nonCtrlSelList.Count > 0) && ((ctrlSelList.Count > 0) || (ctrlUnSelList.Count > 0)))
                                        {
                                            errMsg = "Bad Checkbox Lasso:\n\nYou cannot lasso both Checkbox Controlled checkboxes [the colored checkboxes] and uncontrolled checkboxes at the same time.\n\nAborting selection.";
                                        }
                                    }

                                    if (errMsg.Length > 0)
                                    {
                                        using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                                        {
                                            mainMsg.ShowDialog(this);
                                        }

                                        lassoOK = false;

                                        break;
                                    }
                                }
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.ToString(), "ProcessLeftMouseUpChkBoxCtrl():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        lassoOK = false;
                    }

                    if (lassoOK == true)
                    {
                        // Safe to proceed...so far:
                        if (nonCtrlSelList.Count > 0)
                        {
                            boxSel.cbcSelections = cbcSelection.selNonCtrlBox;
                        }
                        else if (ctrlSelList.Count > 0)
                        {
                            /*
                             * Only one box of this type can be lassoed at once.
                             * 
                             * The controlled checkbox whose associated set will
                             * be selected will be in the boxSel.cbcBoxSetsList
                             * since we maintain a 1:1 relationship with Checkbox
                             * Controls 'Check Box Sets' list.
                             */
                            for (int j = 0; j < boxSel.cbcBoxSetsList.Count; j++)
                            {
                                if (boxSel.cbcBoxSetsList[j].Contains(ctrlSelList[0]))
                                {
                                    //MessageBox.Show(string.Format("and the checkbox is: {0}", field.ref_name));
                                    boxIndex = j;
                                    boxSel.cbcSelections = cbcSelection.selCtrlBoxSet;

                                    break;
                                }
                            }

                            if (boxIndex == -1)
                            {
                                // Error...hmmmmm
                                errMsg = string.Format("ProcessLeftMouseChkBoxCtrlClick() Unexpected Error:\n\n{0}  Could not be located in Calc Helpers copy of Checkbox Controls 'Check Box Sets'. This box will be ignored..sorry.", ctrlSelList[0]);

                                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                                {
                                    mainMsg.ShowDialog(this);
                                }

                                lassoOK = false;
                            }
                        }
                        else if (ctrlUnSelList.Count > 0)
                        {
                            /*
                             * Only one box of this type can be lassoed at once.
                             */
                            boxSel.cbcSelections = cbcSelection.unSelCtrlBox;
                        }
                        else
                        {
                            lassoOK = false;
                        }
                    }
                }

                pdf.mLassoMove = false;
                pdf.mLassoDown = false;

                if (lassoOK == true)
                {
                    switch (boxSel.cbcSelections)
                    {
                        case (cbcSelection.selCtrlBoxSet):

                            if (boxIndex > -1)
                            {
                                ProcessSelectedIndex(boxIndex, "CH");
                            }

                            break;

                        case (cbcSelection.selNonCtrlBox):

                            ProcessSelectedNonCtrlBox(nonCtrlSelList, true, ref pdf);

                            break;

                        case (cbcSelection.unSelCtrlBox):

                            ProcessCtrlBoxUnsel(ctrlUnSelList);

                            break;
                    }
                }
                else
                {
                    // Killing the lasso rectangle:
                    pdf.pictureBox.Refresh();
                }
            }
        }

        /*
         * This routine handles Checkbox Control left mouse click
         * box selections.
         * 
         * Called By:
         * 
         *  pictureBox1_MouseClick()
         *  pictureBox2_MouseClick()
         */
        public void ProcessLeftMouseChkBoxCtrlClick(ref Pdf pdf, ref PageArray pa, Point location)
        {
            //MessageBox.Show("Entered ProcessLeftMouseChkBoxCtrlClick():");

            /*
             * Caller has ensured that pdf, pa and ctClient are not null,
             * and pdf.mLassoMove is false, and it is a left mouse button event.
             * Caller has also checked if Calc Helper is sending checkbox
             * selections to Checkbox Control.
            */
            pdf.mLassoDown = false;

            List<Field> fa = null;
            bool selecting = true;
            bool selOK = true;
            int paIndex = (int)(pdf.curPage - 1);
            int boxIndex = -1;
            List <string> boxList = new List<string>();
            string errMsg = "";

            pa.GetFieldArrayRef(paIndex, ref fa);

            if (fa != null)
            {
                try
                {
                    for (int i = 0; i < fa.Count; i++)
                    {
                        if (fa[i].sbox.Contains(location))
                        {
                            /*
                             * Looking for a checkbox.....
                             */
                            Field field = fa[i];

                            if (field is TextField)
                            {
                                if (((TextField)field).checkbox == true)
                                {
                                    // Are we selecting or unselecting this checkbox?
                                    selecting = !field.leftBoxSel;

                                    if (selecting == true)
                                    {
                                        //MessageBox.Show("We are selecting a checkbox..");

                                        /*
                                         * Is it a noncontrolled checkbox?
                                         */
                                        if (field.ctrlType == -1)
                                        {
                                            boxSel.cbcSelections = cbcSelection.selNonCtrlBox;
                                            boxList.Add(field.ref_name);
                                        }
                                        else if (field.ctrlType == docFile.chkBoxCtrl)
                                        {
                                            /*
                                             * Controlled: BUT - we needed to make sure this is
                                             * a Checkbox Controlled checkbox, not a dataflow box
                                             * for example. field.ctrlType == 0 is Checkbox
                                             * Controlled. We ignore any checkboxes that are not
                                             * Checkbox Controlled.
                                             * 
                                             * The checkbox will be in the boxSel.cbcBoxSetsList
                                             * since we maintain a 1:1 relationship with Checkbox
                                             * Controls 'Check Box Sets' list.
                                             */
                                            for (int j = 0; j < boxSel.cbcBoxSetsList.Count; j++)
                                            {
                                                if (boxSel.cbcBoxSetsList[j].Contains(field.ref_name))
                                                {
                                                    //MessageBox.Show(string.Format("and the checkbox is: {0}", field.ref_name));
                                                    boxIndex = j;
                                                    boxSel.cbcSelections = cbcSelection.selCtrlBoxSet;

                                                    break;
                                                }
                                            }

                                            if (boxIndex == -1)
                                            {
                                                // Error...hmmmmm
                                                errMsg = string.Format("ProcessLeftMouseChkBoxCtrlClick() Unexpected Error:\n\n{0}  Could not be located in Calc Helpers copy of Checkbox Controls 'Check Box Sets'. This box will be ignored..sorry.", field.ref_name);
                                                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                                                {
                                                    mainMsg.ShowDialog(this);
                                                }

                                                selOK = false;
                                            }
                                        }
                                        else
                                        {
                                            selOK = false;
                                        }
                                    }
                                    else
                                    {
                                        /*
                                         * Unselecting a checkbox. Only controlled checkboxes
                                         * are selected, by design. If somebody clicked on a
                                         * currently uncontrolled checkbox we would have selected
                                         * it and added control to it and added it to the current
                                         * checkbox set they were editing.
                                         */
                                        boxSel.cbcSelections = cbcSelection.unSelCtrlBox;
                                        boxList.Add(field.ref_name);
                                    }

                                    break; // for (int i = 0; i < fa.Count; i++) 
                                }
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString(), "ProcessLeftMouseChkBoxCtrlClick():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    selOK = false;
                }

                if (selOK == true)
                {
                    switch (boxSel.cbcSelections)
                    {
                        case (cbcSelection.selCtrlBoxSet):

                            if (boxIndex > -1)
                            {
                                ProcessSelectedIndex(boxIndex, "CH");
                            }

                            break;

                        case (cbcSelection.selNonCtrlBox):

                            ProcessSelectedNonCtrlBox(boxList, false, ref pdf);

                            break;

                        case (cbcSelection.unSelCtrlBox):

                            ProcessCtrlBoxUnsel(boxList);

                            break;
                    }
                }
            }
        }

        /*
         * This routine unselects a controlled checkbox and
         * also removes control from it. The box is removed
         * from the current checkbox set we are editing
         * and we send the edited set to Check Box Control
         * so it can reflect this change to it's setList.
         * 
         * Called From:
         * 
         * Methods3.cs-ProcessLeftMouseChkBoxCtrlClick()
         * Methods3.cs-ProcessLeftMouseUpChkBoxCtrl()
         * 
         * Caller/Caller's Caller has ensured that pa1, pdf1,
         * and ctClient are not null.
         */
        public void ProcessCtrlBoxUnsel(List<string> boxList)
        {
            if (boxList.Count == 1)
            {
                List<string> curBoxSet = new List<string>();
                bool setDeleted = false;
                string boxStr = "";

                try
                {
                    /*
                     * This should be true...
                     */
                    if (boxSel.cbcCurIndex > -1)
                    {
                        boxSel.unselLeftBoxes.Add(boxList[0]);
                        boxSel.ResetSelections(boxSelTypes.leftSel, ref pa1, ref pdf1, ref pa2, ref pdf2);

                        // Remove the box from it's set:
                        curBoxSet = boxSel.cbcBoxSetsList[boxSel.cbcCurIndex].Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();

                        if (curBoxSet.Contains(boxList[0]))
                        {
                            curBoxSet.Remove(boxList[0]);

                            if (curBoxSet.Count == 0)
                            {
                                setDeleted = true;
                                boxSel.cbcBoxSetsList.RemoveAt(boxSel.cbcCurIndex);
                            }
                            else
                            {
                                boxStr = string.Join(",", curBoxSet.ToArray().Where(s => !String.IsNullOrEmpty(s)));
                                boxSel.cbcBoxSetsList[boxSel.cbcCurIndex] = boxStr;
                            }
                        }

                        // Remove the control status:
                        docFile.ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref boxList, docFile.chkBoxCtrl, false);
                        RefreshPanels();

                        // Inform Checkbox Control:

                        if (setDeleted == true)
                        {
                            //MessageBox.Show(string.Format("Tell CBC: Set deleted - cbcBoxSelTypes.delBoxSet\nboxSel.cbcCurIndex = {0}", boxSel.cbcCurIndex));

                            ctClient.SendBoxSelections(string.Format("{0}|{1}", cbcBoxSelTypes.delBoxSet, boxSel.cbcCurIndex));
                            boxSel.cbcCurIndex = -1;
                        }
                        else
                        {
                            string boxSet = string.Format("[{0}]", boxSel.cbcBoxSetsList[boxSel.cbcCurIndex]);

                            ctClient.SendBoxSelections(string.Format("{0}|{1}", cbcBoxSelTypes.updateSel, boxSet));
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString(), "ProcessCtrlBoxUnsel():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        /*
         * This routine processes the selection of one or
         * more noncontrolled checkboxes.
         * 
         * We will set these boxes to the 'controlled' state
         * and insert them into Calc Helpers copy of Checkbox
         * Control's 'Check Box Sets' list at the current
         * list index which both tools maintain. If the
         * current index is -1 we treat this as a new set.
         * 
         * We also make sure if these boxes are going to be
         * inserted into an existing box set that they are 
         * on the same document page. Nothing stopping the
         * user from going to a different page and start
         * selecting uncontrolled checkboxes. Our page numbers
         * are on a zero based index basis, so error messages
         * to the user regarding page errors need to be 
         * incremented by 1 to 'make sense' per the document
         * page.
         * 
         * After all the control and selection processing is
         * done we will refresh the page to reflect the changes.
         * 
         * One box sets introduced this way will be indicated
         * in red in the Checkbox Control tool. User will (and
         * always has been) informed of single box sets and the
         * Copy To Clipboard function in Checkbox Control will
         * refuse to copy under existing single box set presence.
         * 
         * Called From:
         * 
         *      Methods3.cs-ProcessLeftMouseUpChkBoxCtrl()
         *      Methods3.cs-ProcessLeftMouseChkBoxCtrlClick()
         * 
         * Caller/Caller's Caller has ensured that pa1, pdf1,
         * and ctClient are not null.
         */
        public void ProcessSelectedNonCtrlBox(List <string> boxList, bool lasso, ref Pdf pdf)
        {
            int faIndex;
            string errMsg = "";
            bool split = false;
            bool error = false;
            bool newSet = false;
            List<Field> fa1 = null;
            List<Field> fa2 = null;
            List<LeftBoxSel> lbsList = new List<LeftBoxSel>();
            LeftBoxSel lbs;

            try
            {
                // Idiot check - make sure we have boxes!
                if (boxList.Count > 0)
                {
                    if (boxSel.cbcCurIndex == -1)
                    {
                        /*
                         * New Set:
                         * 
                         * Either the Checkbox Sets for Calc Helper
                         * and Checkbox Control are both empty or: the
                         * user has clicked the 'New Set' button on
                         * Checkbox Control, or they just started doing
                         * Checkbox Control and went right in to box
                         * selecting for a new set.
                         */
                        boxSel.cbcBoxSetsList.Add("");
                        boxSel.cbcCurIndex = boxSel.cbcBoxSetsList.Count - 1;

                        /*
                         * In this case the boxSel.cbcCurPage is the page
                         * they are clicking the box on.
                         */
                        boxSel.cbcCurPage = pa1.GetBoxPageNum(boxList[0]) - 1;
                        newSet = true;
                    }

                    /*
                     * Verify page: If a lasso of > 1 box it suffices
                     * to just check the first box since they are all 
                     * on the same page.
                     */
                    int boxPage = pa1.GetBoxPageNum(boxList[0]) - 1;

                    if (boxPage == boxSel.cbcCurPage)
                    {
                        /*
                         * We will need to check if pa2 and pdf2 are null as
                         * we may not have a split screen. If the screen is split
                         * we will need to set the same fields in pa2.
                         */
                        if ((pdf2 != null) && (pa2 != null))
                        {
                            split = true;
                        }

                        if (pdf1.refreshPage == false)
                        {
                            if (pdf1.curPage == (boxPage + 1))
                            {
                                pdf1.refreshPage = true;
                            }
                        }

                        if ((split == true) && (pdf2.refreshPage == false))
                        {
                            if (pdf2.curPage == (boxPage + 1))
                            {
                                pdf2.refreshPage = true;
                            }
                        }

                        // Now set the control to Checkbox Control:
                        pa1.GetFieldArrayRef(boxPage, ref fa1);

                        if (fa1 == null)
                        {
                            using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessSelectedNonCtrlBox() Error:\n\npa1.GetFieldArrayRef() returned null field\narray while processing page {0}.", boxPage)))
                            {
                                mainMsg.ShowDialog(Form1.ActiveForm);
                            }

                            error = true;
                        }

                        if (split == true)
                        {
                            pa2.GetFieldArrayRef(boxPage, ref fa2);

                            if (fa2 == null)
                            {
                                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessSelectedNonCtrlBox() Error:\n\npa2.GetFieldArrayRef() returned null field\narray while processing page {0}.\n\nSelection changes will not be displayed for\nthe second display screen.", boxPage)))
                                {
                                    mainMsg.ShowDialog(Form1.ActiveForm);
                                }

                                pdf2.refreshPage = false;
                                split = false;
                            }
                        }

                        if (error == false)
                        {
                            foreach (string box in boxList)
                            {
                                /*
                                 * NOTE: Since we are setting both the control and
                                 * the left box selection fields on a box I am not 
                                 * calling/using the existing routine in
                                 * ProcessSelections.cs-ProcessLeftBoxSel() to handle
                                 * the left selections. That logic is done here no
                                 * need in going thru all of the fa1/fa2 stuff and
                                 * the split and refresh, etc.. twice - once for the
                                 * control setting and once for the left box selection
                                 * setting.
                                 * 
                                 * Insert each box into the current checkbox set
                                 * and set the box's controlled and selection fields.
                                 */
                                faIndex = pa1.GetBoxFieldArrayIndex(box);

                                lbs = new LeftBoxSel(box, boxPage, faIndex);
                                boxSel.leftBoxSelDict.Add(lbs.box, lbs);

                                boxSel.cbcBoxSetsList[boxSel.cbcCurIndex] += ',' + box;

                                fa1[faIndex].ctrlType = docFile.chkBoxCtrl;
                                fa1[faIndex].leftBoxSel = true;

                                if (split == true)
                                {
                                    /*
                                     * By definition the pa's and fa's are identical since
                                     * they are both tied to the same document.
                                    */
                                    fa2[faIndex].ctrlType = docFile.chkBoxCtrl;
                                    fa2[faIndex].leftBoxSel = true;
                                }
                            }

                            boxSel.cbcBoxSetsList[boxSel.cbcCurIndex] = boxSel.cbcBoxSetsList[boxSel.cbcCurIndex].TrimStart(',');
                            RefreshPanels();
                        }
                    }
                    else
                    {
                        /*
                         * We adjust the page numbers to reflect the user's document perception,
                         * would be confusing if they are currently on document page 3 and see an
                         * error message mentioning [zero index based] page 2 for example.
                         */
                        errMsg = string.Format("Box Selection Page Error\n\nSelected boxes must be on the same page\nas the box set you are editing.\n\n{0} is on page {1} but the checkbox set is on page {2}.\n\nAborting selection.", boxList[0], boxPage + 1, boxSel.cbcCurPage + 1);

                        using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                        {
                            mainMsg.ShowDialog(this);
                        }

                        error = true;

                        if (lasso == true)
                        {
                            /*
                             * Killing the lasso rectangle. There is no
                             * selection rectangle for a bad click.
                             */
                            pdf.pictureBox.Refresh();
                        }

                    }

                    if (error == false)
                    {
                        // Inform Checkbox Control:
                        string boxSet = string.Format("[{0}]", boxSel.cbcBoxSetsList[boxSel.cbcCurIndex]);

                        if (newSet == true)
                        {
                            ctClient.SendBoxSelections(string.Format("{0}|{1}", cbcBoxSelTypes.newCHSet, boxSet));
                        }
                        else
                        {
                            ctClient.SendBoxSelections(string.Format("{0}|{1}", cbcBoxSelTypes.updateSel, boxSet));
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessSelectedNonCtrlBox():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * This routine processes a new existing controlled
         * checkbox set selection. This is either initiated
         * at the Calc Helper end or the Checkbox Control
         * end. We may also setup for a new box set creation
         * if selIndex = -1.
         * 
         * We first check to see if we have a set currently
         * selected. If so we unselect it.
         * 
         * If Checkbox Control initiated this:
         * 
         * If selIndex is -1 then the user wants to start
         * creating a new set. If >= 0 then they selected
         * a new set to edit.
         * 
         * Editing a new set initiated by Checkbox Control:
         * 
         * We then check if the selected set resides on our
         * current page, we also check if we have a split
         * screen and do the same check if true. If on a
         * different page we go to the page so user can see
         * the set and be able to perform edits on it.
         * 
         * We then select the new controlled set.
         * 
         * If Calc Helper initiated this we inform Check Box
         * control so it can select it's set, keeping things
         * 1:1 between the two tools.
         * 
         * Called From:
         * 
         *      Form1.cs-ctClient_CtClientEvent(): user is "CBC"
         *      
         *      Methods3.cs-ProcessLeftMouseChkBoxCtrlClick():
         *          user is "CH"
         * 
         * Caller/Caller's Caller has ensured that pa1, pdf1,
         * and ctClient are not null.
         */
        public void ProcessSelectedIndex(int selIndex, string user)
        {
            //MessageBox.Show(this, string.Format("Entered ProcessSelectedIndex():\n\nselIndex = {0}\n\nuser = {1}\n\nboxSel.cbcCurIndex = {2}\n\nboxSel.cbcBoxSetsList.Count = {3}", selIndex, user, boxSel.cbcCurIndex, boxSel.cbcBoxSetsList.Count));

            List<string> unselectBoxList = new List<string>();
            List<string> selectBoxList = new List<string>();
            List<LeftBoxSel> boxSelList = new List<LeftBoxSel>();
            LeftBoxSel lbs;

            try
            {
                // Unselect the current set's checkboxes if any:
                if (boxSel.cbcCurIndex > -1)
                {
                    unselectBoxList = boxSel.cbcBoxSetsList[boxSel.cbcCurIndex].Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();

                    foreach (string box in unselectBoxList)
                    {
                        boxSel.unselLeftBoxes.Add(box);
                    }

                    boxSel.ResetSelections(boxSelTypes.leftSel, ref pa1, ref pdf1, ref pa2, ref pdf2);
                    RefreshPanels();
                }

                /*
                 * User wants to create a new set?
                 */
                if (selIndex == -1)
                {
                    boxSel.cbcCurIndex = -1;
                }
                else
                {
                    /*
                     * If Checkbox Control initiated this:
                     * 
                     * We first check if we need to change the current
                     * page. We only need to check the first box in the
                     * set as we do not allow the user to build a
                     * checkbox set from more than one page.
                     * 
                     * Next we will select the boxes in this set.
                     * 
                     * The ONLY box on the document that we do not keep track of is
                     * box1_65535, so we can just ask pa1 for the page number here.
                     */
                    selectBoxList = boxSel.cbcBoxSetsList[selIndex].Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();

                    if (user == "CBC")
                    {
                        int boxSetPage = pa1.GetBoxPageNum(selectBoxList[0]);

                        ProcessCBCNewPage(boxSetPage);
                    }

                    foreach (string box in selectBoxList)
                    {
                        lbs = new LeftBoxSel(box, pa1.GetBoxPageNum(box) - 1, pa1.GetBoxFieldArrayIndex(box));
                        boxSelList.Add(lbs);
                    }

                    if (boxSelList.Count > 0)
                    {
                        if (boxSel.ProcessLeftBoxSel(ref boxSelList, ref pdf1, ref pa1, ref pdf2, ref pa2) == true)
                        {
                            RefreshPanels();
                        }

                        boxSel.cbcCurIndex = selIndex;
                        boxSel.cbcCurPage = pa1.GetBoxPageNum(boxSelList[0].box) - 1;

                        if (user == "CH")
                        {
                            // Tell Checkbox Control:
                            ctClient.SendBoxSelections(string.Format("{0}|{1}", cbcBoxSelTypes.newCHSelIndex, selIndex));
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessSelectedIndex():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * This checks if we need to go to the checkbox
         * page belonging to the Checkbox Control box
         * set selection. Obviously the user will need
         * to see the checkboxes so they can perform
         * edits on the set.
         *
         * Called From:
         * 
         *      Methods3.cs-ProcessSelectedIndex()
         * 
         * Caller/Caller's Caller has ensured that pa1, pdf1
         * are not null.
         */
        public void ProcessCBCNewPage(int cbcPage)
        {
            try
            {
                /*
                 * When Calc Helper switched to sending checkbox selections
                 * to Check Box Control it set up pdf1 [and pdf2 if there was
                 * a split screen at the time] for navigating to just pages
                 * with checkboxes.
                 * 
                 * Just checking if they switched out of Checkbox page
                 * navigation - I don't think this is bad to set it back
                 * given the utility in choosing checkboxes by simply clicking
                 * or lassoing them - much easier to go the the next/previous
                 * checkbox page.
                 */
                if (pdf1.pageNav != Pdf.pageNavMode.checkbox)
                {
                    pdf1.pageNav = Pdf.pageNavMode.checkbox;

                    pdf1.boxedPageList = pa1.GetBoxedPageList(pdf1.pageNav.ToString());

                    ProcessPageNavMode(pdf1.pageNav, pdf1.id);
                }

                if (pdf1.curPage != (uint)cbcPage)
                {
                    pdf1.boxedPageIndex = pdf1.boxedPageList.IndexOf((uint)cbcPage);
                    pdf1.curPage = pdf1.boxedPageList[pdf1.boxedPageIndex];
                    pdf1.ShowPage(ref docFile.hPdf, (int)pdf1.curPage, true, ref pa1);

                    ProcessPageNum(ref pdf1);
                }

                if ((splitScreen == true) && (pdf2 != null) && (pa2 != null) && (chkboxToolStripMenuItem2.Enabled == true))
                {
                    if (pdf2.pageNav != Pdf.pageNavMode.checkbox)
                    {
                        /*
                         * They must have split the screen after Checkbox
                         * Control requested box selections from Calc Helper.
                         */
                        pdf2.pageNav = Pdf.pageNavMode.checkbox;

                        /*
                         * Switching to a boxed page navigation mode. Made sure these lists
                         * of page numbers by type have page numbers in them, and the various
                         * tool strip menu items for selecting a particular boxed page
                         * navigation have only been enabled if there are pages with those
                         * boxes on them.
                         */
                        pdf2.boxedPageList = pa2.GetBoxedPageList(pdf2.pageNav.ToString());

                        ProcessPageNavMode(pdf2.pageNav, pdf2.id);
                    }

                    if (pdf2.curPage != (uint)cbcPage)
                    {
                        pdf2.boxedPageIndex = pdf2.boxedPageList.IndexOf((uint)cbcPage);
                        pdf2.curPage = pdf2.boxedPageList[pdf2.boxedPageIndex];
                        pdf2.ShowPage(ref docFile.hPdf, (int)pdf2.curPage, true, ref pa2);

                        ProcessPageNum(ref pdf2);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessCBCNewPage():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * Checkbox Control initiates this process.
         * 
         * This deletes a set from Calc Helper's copy
         * of the 'Check Box Sets' in Checkbox Control.
         * 
         * A setIndex of -1 means we are deleting all
         * box sets. In this case there will be no
         * controlled checkboxes on the document.
         * 
         * We also unselect any checkbox set that is
         * currently selected which SHOULD be the set
         * whose index matches setIndex. 
         * 
         * We have to check if we are deleting the last set
         * in which case we have an empty set list which we
         * need to be aware of.
         * 
         * Called From:
         * 
         * Form1.cs-ctClient_CtClientEvent(): ctEvent.cbcBoxSets
         * 
         * Caller has ensured that pa1, pdf1 are not null.
         */
        public void ProcessSetDeletion(int setIndex)
        {
            if (boxSel.cbcBoxSetsList.Count > 0)
            {
                string boxStr= "";
                List<string> boxList = new List<string>();

                try
                {
                    if (boxSel.cbcCurIndex > -1)
                    {
                        List<string> unselectBoxList = new List<string>();

                        unselectBoxList = boxSel.cbcBoxSetsList[boxSel.cbcCurIndex].Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();

                        foreach (string box in unselectBoxList)
                        {
                            boxSel.unselLeftBoxes.Add(box);
                        }

                        boxSel.ResetSelections(boxSelTypes.leftSel, ref pa1, ref pdf1, ref pa2, ref pdf2);

                        boxSel.cbcCurIndex = -1;
                    }

                    if (setIndex > -1)
                    {
                        boxList = boxSel.cbcBoxSetsList[setIndex].Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();
                        boxSel.cbcBoxSetsList.RemoveAt(setIndex);
                    }
                    else
                    {
                        // All sets:
                        boxStr = string.Join(",", boxSel.cbcBoxSetsList.ToArray().Where(s => !String.IsNullOrEmpty(s)));
                        boxStr = boxStr.TrimEnd(',');
                        boxList = boxStr.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();
                        boxSel.cbcBoxSetsList.Clear();
                    }

                    // Remove the control status:
                    docFile.ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref boxList, docFile.chkBoxCtrl, false);

                    RefreshPanels();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString(), "ProcessSetDeletion():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        /*
         * We set the checkbox control state list,
         * docFile.cbcStateBoxes, to the boxes in
         * setBoxStr when arg useCurrCBCList is false,
         * used when Checkbox Control is initiating
         * this.
         * 
         * When arg useCurrCBCList is true, Calc Helper
         * is the initiator. We use the current
         * checkbox control state in Calc Helper which
         * we build first.
         * 
         * This is just a precaution to keep Calc
         * Helper's checkbox control status in sync
         * with the actual checkbox sets in the Visual
         * Studio Code file after a refresh operation.
         * 
         * The user may do all sorts of adds/deletes
         * regarding checkbox control while using the
         * Checkbox Control tool but never actually
         * updating the document's code file with these
         * changes. In this case Calc Helper will 'look'
         * one way but after a refresh it may not 'look'
         * like it should if we do not make sure the 
         * control list has been updated.
         * 
         * No problem if the user does [which they SHOULD
         * do] update the code file after using Checkbox
         * Control, but this just covers that chance.
         * 
         * Called From:
         * 
         *      ctClient_CtClientEvent()-ctEvent.cbcBoxSets
         *      ProcessChkBoxCtrlMode()
         */
        public void SetCBCStateList(string setBoxStr, bool useCurrCBCList)
        {
            docFile.cbcStateBoxes.Clear();

            if (useCurrCBCList == true)
            {
                /*
                 * Calc Helper's clone of Checkbox Control's
                 * checkbox setList.
                 */
                if (boxSel.cbcBoxSetsList.Count > 0)
                {
                    string boxStr = "";

                    boxStr = string.Join(",", boxSel.cbcBoxSetsList.ToArray().Where(s => !String.IsNullOrEmpty(s)));
                    boxStr = boxStr.TrimEnd(',');
                    docFile.cbcStateBoxes = boxStr.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();
                }
            }
            else
            {
                // Length == 0 means no boxes.
                if (setBoxStr.Length > 0)
                {
                    docFile.cbcStateBoxes = setBoxStr.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();
                }
            }
        }
    }
}